# general
class AutoControlException(Exception):
    pass


# Keyboard


class AutoControlKeyboardException(AutoControlException):
    pass


class AutoControlCantFindKeyException(AutoControlKeyboardException):
    pass


# Mouse


class AutoControlMouseException(AutoControlException):
    pass


# Screen


class AutoControlScreenException(AutoControlException):
    pass


# Image detect


class ImageNotFoundException(AutoControlException):
    pass


# Record


class AutoControlRecordException(AutoControlException):
    pass


# Execute action

class AutoControlExecuteActionException(AutoControlException):
    pass


class AutoControlJsonActionException(AutoControlExecuteActionException):
    pass


class AutoControlActionNullException(AutoControlExecuteActionException):
    pass


class AutoControlActionException(AutoControlExecuteActionException):
    pass


class AutoControlAddCommandException(AutoControlExecuteActionException):
    pass


class AutoControlArgparseException(AutoControlExecuteActionException):
    pass


# timeout
class AutoControlTimeoutException(AutoControlException):
    pass


# html exception

class AutoControlHTMLException(AutoControlException):
    pass


# Json Exception

class AutoControlJsonException(AutoControlException):
    pass


class AutoControlGenerateJsonReportException(AutoControlJsonException):
    pass


# XML

class XMLException(AutoControlException):
    pass


class XMLTypeException(XMLException):
    pass


# Execute callback
class CallbackExecutorException(AutoControlException):
    pass



import builtins
import types
from inspect import getmembers, isbuiltin
from typing import Any, Dict, List, Union

from je_auto_control.utils.exception.exception_tags import action_is_null_error, add_command_exception, \
    executor_list_error
from je_auto_control.utils.exception.exception_tags import cant_execute_action_error
from je_auto_control.utils.exception.exceptions import AutoControlActionException, AutoControlAddCommandException
from je_auto_control.utils.exception.exceptions import AutoControlActionNullException
from je_auto_control.utils.generate_report.generate_html_report import generate_html
from je_auto_control.utils.generate_report.generate_html_report import generate_html_report
from je_auto_control.utils.generate_report.generate_json_report import generate_json
from je_auto_control.utils.generate_report.generate_json_report import generate_json_report
from je_auto_control.utils.generate_report.generate_xml_report import generate_xml
from je_auto_control.utils.generate_report.generate_xml_report import generate_xml_report
from je_auto_control.utils.json.json_file import read_action_json
from je_auto_control.utils.logging.loggin_instance import auto_control_logger
from je_auto_control.utils.package_manager.package_manager_class import package_manager
from je_auto_control.utils.project.create_project_structure import create_project_dir
from je_auto_control.utils.scheduler.extend_apscheduler import scheduler_manager
from je_auto_control.utils.shell_process.shell_exec import ShellManager
from je_auto_control.utils.start_exe.start_another_process import start_exe
from je_auto_control.utils.test_record.record_test_class import record_action_to_list, test_record_instance
from je_auto_control.wrapper.auto_control_image import locate_all_image, locate_and_click, locate_image_center
from je_auto_control.wrapper.auto_control_keyboard import check_key_is_press
from je_auto_control.wrapper.auto_control_keyboard import get_special_table, get_keyboard_keys_table
from je_auto_control.wrapper.auto_control_keyboard import press_keyboard_key, release_keyboard_key, hotkey, \
    type_keyboard, write
from je_auto_control.wrapper.auto_control_mouse import get_mouse_position, press_mouse, release_mouse, click_mouse, \
    mouse_scroll
from je_auto_control.wrapper.auto_control_mouse import get_mouse_table
from je_auto_control.wrapper.auto_control_mouse import set_mouse_position
from je_auto_control.wrapper.auto_control_record import record, stop_record
from je_auto_control.wrapper.auto_control_screen import screenshot, screen_size


class Executor(object):

    def __init__(self):
        self.event_dict: dict = {
            # mouse
            "AC_mouse_left": click_mouse,
            "AC_mouse_right": click_mouse,
            "AC_mouse_middle": click_mouse,
            "AC_click_mouse": click_mouse,
            "AC_get_mouse_table": get_mouse_table,
            "AC_get_mouse_position": get_mouse_position,
            "AC_press_mouse": press_mouse,
            "AC_release_mouse": release_mouse,
            "AC_mouse_scroll": mouse_scroll,
            "AC_set_mouse_position": set_mouse_position,
            "AC_get_special_table": get_special_table,
            # keyboard
            "AC_get_keyboard_keys_table": get_keyboard_keys_table,
            "AC_type_keyboard": type_keyboard,
            "AC_press_keyboard_key": press_keyboard_key,
            "AC_release_keyboard_key": release_keyboard_key,
            "AC_check_key_is_press": check_key_is_press,
            "AC_write": write,
            "AC_hotkey": hotkey,
            # image
            "AC_locate_all_image": locate_all_image,
            "AC_locate_image_center": locate_image_center,
            "AC_locate_and_click": locate_and_click,
            # screen
            "AC_screen_size": screen_size,
            "AC_screenshot": screenshot,
            # test record
            "AC_set_record_enable": test_record_instance.set_record_enable,
            # only generate
            "AC_generate_html": generate_html,
            "AC_generate_json": generate_json,
            "AC_generate_xml": generate_xml,
            # generate report
            "AC_generate_html_report": generate_html_report,
            "AC_generate_json_report": generate_json_report,
            "AC_generate_xml_report": generate_xml_report,
            # record
            "AC_record": record,
            "AC_stop_record": stop_record,
            # execute
            "AC_execute_action": self.execute_action,
            "AC_execute_files": self.execute_files,
            "AC_add_package_to_executor": package_manager.add_package_to_executor,
            "AC_add_package_to_callback_executor": package_manager.add_package_to_callback_executor,
            # project
            "AC_create_project": create_project_dir,
            # Shell
            "AC_shell_command": ShellManager().exec_shell,
            # Another process
            "AC_execute_process": start_exe,
            # Scheduler
            "AC_scheduler_event_trigger": self.scheduler_event_trigger,
            "AC_remove_blocking_scheduler_job": scheduler_manager.remove_blocking_job,
            "AC_remove_nonblocking_scheduler_job": scheduler_manager.remove_nonblocking_job,
            "AC_start_blocking_scheduler": scheduler_manager.start_block_scheduler,
            "AC_start_nonblocking_scheduler": scheduler_manager.start_nonblocking_scheduler,
            "AC_start_all_scheduler": scheduler_manager.start_all_scheduler,
            "AC_shutdown_blocking_scheduler": scheduler_manager.shutdown_blocking_scheduler,
            "AC_shutdown_nonblocking_scheduler": scheduler_manager.shutdown_nonblocking_scheduler,
        }
        # get all builtin function and add to event dict
        for function in getmembers(builtins, isbuiltin):
            self.event_dict.update({str(function[0]): function[1]})

    def _execute_event(self, action: list) -> Any:
        event = self.event_dict.get(action[0])
        if len(action) == 2:
            if isinstance(action[1], dict):
                return event(**action[1])
            else:
                return event(*action[1])
        elif len(action) == 1:
            return event()
        else:
            raise AutoControlActionException(cant_execute_action_error + " " + str(action))

    def execute_action(self, action_list: [list, dict]) -> Dict[str, str]:
        """
        use to execute all action on action list(action file or program list)
        :param action_list the list include action
        for loop the list and execute action
        """
        auto_control_logger.info(f"execute_action, action_list: {action_list}")
        if isinstance(action_list, dict):
            action_list: list = action_list.get("auto_control", None)
            if action_list is None:
                raise AutoControlActionNullException(executor_list_error)
        execute_record_dict = dict()
        try:
            if len(action_list) > 0 or isinstance(action_list, list):
                pass
            else:
                raise AutoControlActionNullException(action_is_null_error)
        except Exception as error:
            record_action_to_list("AC_execute_action", action_list, repr(error))
            auto_control_logger.info(
                f"execute_action, action_list: {action_list}, "
                f"failed: {repr(error)}")
        for action in action_list:
            try:
                event_response = self._execute_event(action)
                execute_record = "execute: " + str(action)
                execute_record_dict.update({execute_record: event_response})
            except Exception as error:
                auto_control_logger.info(
                    f"execute_action, action_list: {action_list}, "
                    f"action: {action}, failed: {repr(error)}")
                record_action_to_list("AC_execute_action", None, repr(error))
                execute_record = "execute: " + str(action)
                execute_record_dict.update({execute_record: repr(error)})
        for key, value in execute_record_dict.items():
            print(key, flush=True)
            print(value, flush=True)
        return execute_record_dict

    def execute_files(self, execute_files_list: list) -> List[Dict[str, str]]:
        """
        :param execute_files_list: list include execute files path
        :return: every execute detail as list
        """
        auto_control_logger.info(f"execute_files, execute_files_list: {execute_files_list}")
        execute_detail_list: list = list()
        for file in execute_files_list:
            execute_detail_list.append(self.execute_action(read_action_json(file)))
        return execute_detail_list

    def scheduler_event_trigger(
            self, function: str, id: str = None, args: Union[list, tuple] = None,
            kwargs: dict = None, scheduler_type: str = "nonblocking", wait_type: str = "secondly",
            wait_value: int = 1, **trigger_args: Any) -> None:
        if scheduler_type == "nonblocking":
            scheduler_event = scheduler_manager.nonblocking_scheduler_event_dict.get(wait_type)
        else:
            scheduler_event = scheduler_manager.blocking_scheduler_event_dict.get(wait_type)
        scheduler_event(self.event_dict.get(function), id, args, kwargs, wait_value, **trigger_args)


executor = Executor()
package_manager.executor = executor


def add_command_to_executor(command_dict: dict) -> None:
    """
    :param command_dict: dict include command we want to add to event_dict
    """
    for command_name, command in command_dict.items():
        if isinstance(command, (types.MethodType, types.FunctionType)):
            executor.event_dict.update({command_name: command})
        else:
            raise AutoControlAddCommandException(add_command_exception)


def execute_action(action_list: list) -> Dict[str, str]:
    return executor.execute_action(action_list)


def execute_files(execute_files_list: list) -> List[Dict[str, str]]:
    return executor.execute_files(execute_files_list)



import json
from pathlib import Path
from threading import Lock
from typing import List, Dict

from je_auto_control.utils.exception.exception_tags import cant_find_json_error
from je_auto_control.utils.exception.exception_tags import cant_save_json_error
from je_auto_control.utils.exception.exceptions import AutoControlJsonActionException

_lock = Lock()


def read_action_json(json_file_path: str) -> List[List[Dict[str, Dict[str, str]]]]:
    """
    use to read action file
    :param json_file_path json file's path to read
    """
    _lock.acquire()
    try:
        file_path = Path(json_file_path)
        if file_path.exists() and file_path.is_file():
            with open(json_file_path) as read_file:
                return json.loads(read_file.read())
    except AutoControlJsonActionException:
        raise AutoControlJsonActionException(cant_find_json_error)
    finally:
        _lock.release()


def write_action_json(json_save_path: str, action_json: list) -> None:
    """
    use to save action file
    :param json_save_path  json save path
    :param action_json the json str include action to write
    """
    _lock.acquire()
    try:
        with open(json_save_path, "w+") as file_to_write:
            file_to_write.write(json.dumps(action_json, indent=4))
    except AutoControlJsonActionException:
        raise AutoControlJsonActionException(cant_save_json_error)
    finally:
        _lock.release()



import sys

from je_auto_control.utils.exception.exception_tags import linux_import_error
from je_auto_control.utils.exception.exceptions import AutoControlException

if sys.platform not in ["linux", "linux2"]:
    raise AutoControlException(linux_import_error)

import os
from Xlib.display import Display


# get x system display
display = Display(os.environ['DISPLAY'])



import sys

from je_auto_control.utils.exception.exception_tags import windows_import_error
from je_auto_control.utils.exception.exceptions import AutoControlException

if sys.platform not in ["win32", "cygwin", "msys"]:
    raise AutoControlException(windows_import_error)

# windows mouse virtual keycode


win32_MOVE: int = 0x0001
win32_LEFTDOWN: int = 0x0002
win32_LEFTUP: int = 0x0004
win32_RIGHTDOWN: int = 0x0008
win32_RIGHTUP: int = 0x0010
win32_MIDDLEDOWN: int = 0x0020
win32_MIDDLEUP: int = 0x0040
win32_DOWN: int = 0x0080
win32_XUP: int = 0x0100
win32_WHEEL: int = 0x0800
win32_HWHEEL: int = 0x1000
win32_ABSOLUTE: int = 0x8000
win32_XBUTTON1: int = 0x0001
win32_XBUTTON2: int = 0x0002

win32_VK_LBUTTON: int = 0x01
win32_VK_RBUTTON: int = 0x02
win32_VK_MBUTTON: int = 0x04
win32_VK_XBUTTON1: int = 0x05
win32_VK_XBUTTON2: int = 0x06

"""
windows keyboard virtual keycode
"""

win32_EventF_EXTENDEDKEY: int = 0x0001
win32_EventF_KEYUP: int = 0x0002
win32_EventF_UNICODE: int = 0x0004
win32_EventF_SCANCODE: int = 0x0008

win32_VkToVSC: int = 0
win32_VK_CANCEL: int = 0x03
win32_VK_BACK: int = 0x08  # BACKSPACE key
win32_VK_TAB: int = 0x09  # TAB key
win32_VK_CLEAR: int = 0x0C  # CLEAR key
win32_VK_RETURN: int = 0x0D  # ENTER key
win32_VK_SHIFT: int = 0x10  # SHIFT key
win32_VK_CONTROL: int = 0x11  # CTRL key
win32_VK_Menu: int = 0x12  # ALT key
win32_VK_PAUSE: int = 0x13  # PAUSE key
win32_VK_CAPITAL: int = 0x14  # CAPS LOCK key
win32_VK_KANA: int = 0x15
win32_VK_IME_ON: int = 0x16
win32_VK_JUNJA: int = 0x17
win32_VK_FINAL: int = 0x18  # ESC key
win32_VK_HANJA: int = 0x19
win32_VK_IME_OFF: int = 0x1A
win32_VK_ESCAPE: int = 0x1B
win32_VK_CONVERT: int = 0x1C
win32_VK_NONCONVERT: int = 0x1D
win32_VK_ACCEPT: int = 0x1E
win32_VK_MODECHANGE: int = 0x1F
win32_VK_SPACE: int = 0x20  # SPACEBAR
win32_VK_PRIOR: int = 0x21  # PAGE UP key
win32_VK_NEXT: int = 0x22  # PAGE DOWN key
win32_VK_END: int = 0x23  # END key
win32_VK_HOME: int = 0x24  # HOME key
win32_VK_LEFT: int = 0x25  # LEFT ARROW key
win32_VK_UP: int = 0x26
win32_VK_RIGHT: int = 0x27
win32_VK_DOWN: int = 0x28
win32_VK_SELECT: int = 0x29
win32_VK_PRINT: int = 0x2A
win32_VK_EXECUTE: int = 0x2B
win32_VK_SNAPSHOT: int = 0x2C
win32_VK_INSERT: int = 0x2D
win32_VK_DELETE: int = 0x2E
win32_VK_HELP: int = 0x2F
win32_key0: int = 0x30
win32_key1: int = 0x31
win32_key2: int = 0x32
win32_key3: int = 0x33
win32_key4: int = 0x34
win32_key5: int = 0x35
win32_key6: int = 0x36
win32_key7: int = 0x37
win32_key8: int = 0x38
win32_key9: int = 0x39
win32_keyA: int = 0x41
win32_keyB: int = 0x42
win32_keyC: int = 0x43
win32_keyD: int = 0x44
win32_keyE: int = 0x45
win32_keyF: int = 0x46
win32_keyG: int = 0x47
win32_keyH: int = 0x48
win32_keyI: int = 0x49
win32_keyJ: int = 0x4A
win32_keyK: int = 0x4B
win32_keyL: int = 0x4C
win32_keyM: int = 0X4D
win32_keyN: int = 0x4E
win32_keyO: int = 0x4F
win32_keyP: int = 0x50
win32_keyQ: int = 0x51
win32_keyR: int = 0x52
win32_keyS: int = 0x53
win32_keyT: int = 0x54
win32_keyU: int = 0x55
win32_keyV: int = 0x56
win32_keyW: int = 0x57
win32_keyX: int = 0x58
win32_keyY: int = 0x59
win32_keyZ: int = 0x5A
win32_VK_LWIN: int = 0x5B  # Left Windows key (Natural keyboard)
win32_VK_RWIN: int = 0x5C  # Right Windows key (Natural keyboard)
win32_VK_APPS: int = 0x5D  # Applications key (Natural keyboard)
win32_VK_SLEEP: int = 0x5F  # Computer Sleep key
win32_VK_NUMPAD0: int = 0x60  # Numeric keypad 0 key
win32_VK_NUMPAD1: int = 0x61
win32_VK_NUMPAD2: int = 0x62
win32_VK_NUMPAD3: int = 0x63
win32_VK_NUMPAD4: int = 0x64
win32_VK_NUMPAD5: int = 0x65
win32_VK_NUMPAD6: int = 0x66
win32_VK_NUMPAD7: int = 0x67
win32_VK_NUMPAD8: int = 0x68
win32_VK_NUMPAD9: int = 0x69
win32_VK_MULTIPLY: int = 0x6A  # Multiply key
win32_VK_ADD: int = 0x6B  # Add key
win32_VK_SEPARATOR: int = 0x6C  # Separator key
win32_VK_SUBTRACT: int = 0x6D  # Subtract key
win32_VK_DECIMAL: int = 0x6E  # Decimal key
win32_VK_DIVIDE: int = 0x6F  # VK_DIVIDE
win32_VK_F1: int = 0x70  # F1
win32_VK_F2: int = 0x71
win32_VK_F3: int = 0x72
win32_VK_F4: int = 0x73
win32_VK_F5: int = 0x74
win32_VK_F6: int = 0x75
win32_VK_F7: int = 0x76
win32_VK_F8: int = 0x77
win32_VK_F9: int = 0x78
win32_VK_F10: int = 0x79
win32_VK_F11: int = 0x7A
win32_VK_F12: int = 0x7B
win32_VK_F13: int = 0x7C
win32_VK_F14: int = 0x7D
win32_VK_F15: int = 0x7E
win32_VK_F16: int = 0x7F
win32_VK_F17: int = 0x80
win32_VK_F18: int = 0x81
win32_VK_F19: int = 0x82
win32_VK_F20: int = 0x83
win32_VK_F21: int = 0x84
win32_VK_F22: int = 0x85
win32_VK_F23: int = 0x86
win32_VK_F24: int = 0x87
win32_VK_NUMLOCK: int = 0x90  # NUM LOCK key
win32_VK_SCROLL: int = 0x91  # SCROLL LOCK key
win32_VK_LSHIFT: int = 0xA0  # Left SHIFT key
win32_VK_RSHIFT: int = 0xA1
win32_VK_LCONTROL: int = 0xA2  # Left CONTROL key
win32_VK_RCONTROL: int = 0xA3  # Right CONTROL key
win32_VK_LMENU: int = 0xA4  # Left MENU key
win32_VK_RMENU: int = 0xA5  # Right MENU key
win32_VK_BROWSER_BACK: int = 0xA6  # Browser Back key
win32_VK_BROWSER_FORWARD: int = 0xA7  # Browser Forward key
win32_VK_BROWSER_REFRESH: int = 0xA8  # Browser Refresh key
win32_VK_BROWSER_STOP: int = 0xA9  # Browser Stop key
win32_VK_BROWSER_SEARCH: int = 0xAA  # Browser Search key
win32_VK_BROWSER_FAVORITES: int = 0xAB  # Browser Favorites key
win32_VK_VOLUME_MUTE: int = 0xAD
win32_VK_VOLUME_DOWN: int = 0xAE
win32_VK_VOLUME_UP: int = 0xAF
win32_VK_MEDIA_NEXT_TRACK: int = 0xB0
win32_VK_MEDIA_PREV_TRACK: int = 0xB1
win32_VK_MEDIA_STOP: int = 0xB2
win32_VK_MEDIA_PLAY_PAUSE: int = 0xB3
win32_VK_LAUNCH_MAIL: int = 0xB4
win32_VK_LAUNCH_MEDIA_SELECT: int = 0xB5
win32_VK_LAUNCH_APP1: int = 0xB6
win32_VK_LAUNCH_APP2: int = 0xB7



from os import getcwd
from pathlib import Path
from threading import Lock

from je_auto_control.utils.json.json_file import write_action_json
from je_auto_control.utils.logging.loggin_instance import auto_control_logger
from je_auto_control.utils.project.template.template_executor import executor_template_1, \
    executor_template_2, bad_executor_template_1
from je_auto_control.utils.project.template.template_keyword import template_keyword_1, \
    template_keyword_2, bad_template_1


def create_dir(dir_name: str) -> None:
    """
    :param dir_name: create dir use dir name
    :return: None
    """
    Path(dir_name).mkdir(
        parents=True,
        exist_ok=True
    )


def create_template(parent_name: str, project_path: str = None) -> None:
    if project_path is None:
        project_path = getcwd()
    keyword_dir_path = Path(project_path + "/" + parent_name + "/keyword")
    executor_dir_path = Path(project_path + "/" + parent_name + "/executor")
    lock = Lock()
    if keyword_dir_path.exists() and keyword_dir_path.is_dir():
        write_action_json(project_path + "/" + parent_name + "/keyword/keyword1.json", template_keyword_1)
        write_action_json(project_path + "/" + parent_name + "/keyword/keyword2.json", template_keyword_2)
        write_action_json(project_path + "/" + parent_name + "/keyword/bad_keyword_1.json", bad_template_1)
    if executor_dir_path.exists() and keyword_dir_path.is_dir():
        lock.acquire()
        try:
            with open(project_path + "/" + parent_name + "/executor/executor_one_file.py", "w+") as file:
                file.write(
                    executor_template_1.replace(
                        "{temp}",
                        project_path + "/" + parent_name + "/keyword/keyword1.json"
                    )
                )
            with open(project_path + "/" + parent_name + "/executor/executor_bad_file.py", "w+") as file:
                file.write(
                    bad_executor_template_1.replace(
                        "{temp}",
                        project_path + "/" + parent_name + "/keyword/bad_keyword_1.json"
                    )
                )
            with open(project_path + "/" + parent_name + "/executor/executor_folder.py", "w+") as file:
                file.write(
                    executor_template_2.replace(
                        "{temp}",
                        project_path + "/" + parent_name + "/keyword"
                    )
                )
        finally:
            lock.release()


def create_project_dir(project_path: str = None, parent_name: str = "AutoControl") -> None:
    auto_control_logger.info(f"create_project_dir, project_path: {project_path}, parent_name: {parent_name}")
    if project_path is None:
        project_path = getcwd()
    create_dir(project_path + "/" + parent_name + "/keyword")
    create_dir(project_path + "/" + parent_name + "/executor")
    create_template(parent_name)



import _thread
import sys
from threading import Thread

from je_auto_control.wrapper.auto_control_keyboard import keyboard_keys_table
from je_auto_control.wrapper.platform_wrapper import keyboard_check


class CriticalExit(Thread):
    """
    use to make program interrupt
    """

    def __init__(self, default_daemon: bool = True):
        """
        default interrupt is keyboard F7 key
        :param default_daemon bool thread setDaemon
        """
        super().__init__()
        self.daemon = default_daemon
        self._exit_check_key: int = keyboard_keys_table.get("f7")

    def set_critical_key(self, keycode: [int, str] = None) -> None:
        """
        set interrupt key
        :param keycode interrupt key
        """
        if isinstance(keycode, int):
            self._exit_check_key = keycode
        else:
            self._exit_check_key = keyboard_keys_table.get(keycode)

    def run(self) -> None:
        """
        listener keycode _exit_check_key to interrupt
        """
        try:
            while True:
                if keyboard_check.check_key_is_press(self._exit_check_key):
                    _thread.interrupt_main()
        except Exception as error:
            print(repr(error), file=sys.stderr)

    def init_critical_exit(self) -> None:
        """
        should only use this to start critical exit
        may this function will add more
        """
        critical_thread = self
        critical_thread.start()
